#!/usr/bin/env -S 64tass --m6502 --ascii -o softwired.prg

; Copyright 2021 Gunnar Ruthenberg and Ronny Engmann

;    This program is free software: you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation, either version 3 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program.  If not, see <https://www.gnu.org/licenses/>.

;    The music is licensed under
;    Creative Commons Attribution + ShareAlike CC BY-SA 4.0.


; full framerate hires "vector" font infinite zoom scroller

NTSC              = 1

; run-time memory is $0000 up to at most $13d9 incl.,
; $0000..$01ff and $0300..$0333 are not counted,
; counting as $03da..$13d9 as per ICC 4K rules,
; with $0200..$02ff, $0334..$03d9 and memory at $1400+
; being buffered at start within the $03da..$13d9 range and restored upon exit

LOWMEM0           = $0200; $0100 bytes
LOWMEM1           = $0334; $a6 bytes

LOWMEM0BUFFER     = $0404
LOWMEM0BUFFERSIZE = $0504 - LOWMEM0BUFFER
LOWMEMGAPSIZE     = $34; $0300..$0333 is not counted
LOWMEM1BUFFERSIZE = lowmemcodeend - logosprite

PAYLOAD           = LOWMEM0 + $1000 + LOWMEM0BUFFERSIZE + LOWMEMGAPSIZE + LOWMEM1BUFFERSIZE


FG                = 7
BG                = 6

.if NTSC
SCREENSHRINK      = 53
SCREENSHRINKMOVE  = 5
CYCLESPERLINE     = 65
LOGOROWOFFSET     = 4 * 40
MAXROWS           = 21
COLUMNS           = 57

TEXELASPECTRATIO  = 1.01 * 256 * 7 / (3 * 8)
.else
SCREENSHRINK      = 20
SCREENSHRINKMOVE  = 6
CYCLESPERLINE     = 63
LOGOROWOFFSET     = 0
MAXROWS           = 25
COLUMNS           = 63

TEXELASPECTRATIO  = 1.1 * 256 * 7 / (3 * 8)
.endif

SOFTSCROLLSPEED   = $40
ZOOMSPEED         = $0480
SLOWZOOMSPEED     = $20
INITIALZOOM       = $30
INITIALZOOMSPEED  = -$80
MAXZOOM           = 192

LOOPTOS           = 13
CONST0F           = $83

STACKPOINTER      = $02
CHARBUFFERPTR     = $03
LEFTTABLE         = $05
RIGHTTABLE        = $09
POINTER           = $0d
COUNTER           = POINTER
HARDSCROLLHI      = $0f
SLOWZOOM          = $10
SCALE             = $11
ZOOMROWMAPPING    = $13

MUSICZP           = $16
CHARCOLUMN        = MUSICZP1
SCROLLY           = MUSICZP1 + 7

SCRATCH           = $6f
ROWCOUNT          = SCRATCH
SOURCE            = SCRATCH
CENTER            = SCRATCH
SCRATCHHI         = $77
SOFTSCROLL        = SCRATCHHI
ROWHEIGHT         = $7f
CHARPTR           = $87
COLUMNCOUNT       = $8f
ZOOMSRCLO         = $97
ZOOMSRCHI         = $9f
CENTERHI          = $a7
ZOOMCOUNT         = $af
YSCALELO          = $b7
YSCALEHI          = $bf
ZOOMSWITCH        = $c7
REMAPSWITCH       = $cf
HARDSCROLLLO      = $d7
YSCALESWITCH      = $df
SCROLLPOS         = $e7
SCROLL            = $ef
SCROLLSEMAPHORE   = $f7

CHARSET           = $0000
YSCROLLTOP        = $0240 - MAXROWS
XSCROLL           = $0340 - MAXROWS
CHARSCRATCH       = $0504
YSCROLLBOTTOM     = $0600 - MAXROWS
CHARBUFFER        = $0600

COLUMNSELECT      = $1500

CHARBUFNUMROWS    = 16
CHARBUFNUMPAGES   = 3

                * = $0801

LINENUMBER        = 2021

                .byte <zero, >zero, <LINENUMBER, >LINENUMBER, $9e
                .text format("%d", start)
zero            .byte 0

start           lda #0
                sta $d020
                sta $d021
                jsr $e536

                lda #$7f
                sta $dc00
                sta $dc0d

                dex
-               lda LOWMEM0,x
                sta LOWMEM0BUFFER,x
                cpx #LOWMEM1BUFFERSIZE
                bcs +
                lda logosprite,x
                sta $0540,x
                lda LOWMEM1,x
                sta logosprite,x
+               inx
                bne -

char .function arg
     .endf arg * 8

cset .function dest
     .endf char(char(dest)) - 1

makechars       ldy #8
makechars0      lda #$ff
                sta SCRATCH
makechars1      lda #$ff
-               sta CHARSET + cset(%0101),x
                lsr SCRATCH
                ror
                inx
                dey
                bne -
                lsr makechars0 + 1
                ror makechars1 + 1
                txa
                bpl makechars

                ldx #8 * 8
copychars
                EMPTY            = %0011
                FULL             = %0111
                EMPTYTOFULLLEFT  = %1010
                EMPTYTOFULLRIGHT = %1101
                FULLTOEMPTYLEFT  = %1110
                FULLTOEMPTYRIGHT = %1001

top    .function arg
       .endf arg | %1000

bottom .function arg
       .endf arg & %0111

                lda CHARSET + cset(bottom(EMPTYTOFULLRIGHT)),x
                sta CHARSET + cset(top(FULL)),x
                eor #$ff
                sta CHARSET + cset(bottom(FULLTOEMPTYRIGHT)),x
                sta CHARSET + cset(top(EMPTY)),x

                lda CHARSET + cset(bottom(FULLTOEMPTYLEFT)),x
                sta CHARSET + cset(bottom(EMPTYTOFULLLEFT)),x
                sta CHARSET + cset(EMPTYTOFULLRIGHT),x
                eor #$ff
                sta CHARSET + cset(bottom(FULLTOEMPTYLEFT)),x
                sta CHARSET + cset(FULLTOEMPTYRIGHT),x
                tya
                sta $02,x
                cpx #8
                bcs +
                sta CHARSCRATCH,x
+               sta CHARSET + cset(EMPTY),x
                sta CHARSET + cset(EMPTYTOFULLLEFT),x
                lda #$ff
                sta CHARSET + cset(FULL),x
                sta CHARSET + cset(FULLTOEMPTYLEFT),x
                dex
                bne copychars

                ldx #$1b
-               lda initvals - 1,x
                sta $fffa - 1,x
                dex
                bne -

                lda #'+'
                sta $0402 + LOGOROWOFFSET
                lda #'h'
                sta $0403 + LOGOROWOFFSET
.if SCREENSHRINK >= 50
                lda #$34 + SCREENSHRINK
.else
                lda #$2e + SCREENSHRINK + ((SCREENSHRINK + 3) / 8)
.endif
                .cerror * != $08a3, "sync not at $08a3"
sync = * + 1
                ldx #$9e
                ldy #8
                cpx #0
                bne sync

                sta $d012

                lax #0
                ldy #CYCLESPERLINE - 1
                sty $dd04
                sta $dd05
                ldy #6
                sty $dd06
                sta $dd07
                lda #%00010101
                ldy #$0b
                sta $dd0e
                sty $d011

                sty $d029
                iny

-               lda #FG
                sta $d918,x
                lda #0
                sta $d800,x
                lda #EMPTYCHAR
.for charbufferpage := 0, charbufferpage < (CHARBUFNUMPAGES - 1), charbufferpage += 1
                sta CHARBUFFER + ($0100 * charbufferpage),x
.next
                inx
                bne -
                beq torun; jmp

                INITIALPOS = CHARBUFFER + (12 * 2 * CHARBUFNUMROWS)
initvals        .word $dd0c, 0, irqhandler
                .byte $2f, $35
                .byte LOOPTOS - 2
                .word INITIALPOS

                EMPTYCHAR             = char(EMPTY)
                FULLCHAR              = char(FULL)
                EMPTYTOFULLLEFTCHAR   = char(EMPTYTOFULLLEFT)
                EMPTYTOFULLLRIGHTCHAR = char(EMPTYTOFULLRIGHT)
                FULLTOEMPTYLEFTCHAR   = char(FULLTOEMPTYLEFT)
                FULLTOEMPTYRIGHTCHAR  = char(FULLTOEMPTYRIGHT)

                .byte EMPTYCHAR, EMPTYTOFULLLEFTCHAR,   FULLTOEMPTYLEFTCHAR,  FULLCHAR
                .byte EMPTYCHAR, EMPTYTOFULLLRIGHTCHAR, FULLTOEMPTYRIGHTCHAR, FULLCHAR

                .word COLUMNSELECT - $0200 + RENDERJUMPSIZE
                .byte >CHARBUFFER
                .char INITIALZOOMSPEED
                .byte 0, INITIALZOOM
                .byte 11, 12

                .cerror * > (CHARBUFFER + ($0100 * CHARBUFNUMPAGES)), "init too large"

                * = CHARBUFFER + ($0100 * CHARBUFNUMPAGES)

                jmp irqout

torun           sty $d802 + LOGOROWOFFSET
                jmp run

RENDERJUMPSIZE = renderjumpend - renderjump

renderjump      ldx #0
                jmp renderchunks.chunk[0]
renderjumpend

RENDERCHUNKSIZE  = 6
CHARBUFFEROFFSET = 1
MASKOFFSET       = 4

renderchunks .bfor column := 1, column <= COLUMNS, column += 1
chunk           lda CHARBUFFER,x
                ora #$00
                pha
.next
                bne *

.if COLUMNS < 60
                .fill 6 * (60 - COLUMNS)
.endif

irqhandler      pha
                txa
                pha
                tya
                pha

                lda #$35
                sta $01

                lda #$ff
                eor $dd04
                asr #%00000111
                sta stabilise + 1
                bcc stabilise
stabilise       bpl *
                nop
                nop
                nop
.if NTSC
                nop
                nop
                nop
.endif
.if SCREENSHRINK >= 50
                clc
                lda #(SCREENSHRINK + ((SCREENSHRINK + 3) / 8) + 1) & 7
                ldy #2
.else
                nop
                lda #(SCREENSHRINK + ((SCREENSHRINK + 3) / 8)) & 7
                ldy #7 - ((SCREENSHRINK + 3) / 8)
.endif
                bne linecrunchentry; jmp

withtoprow      pla
                pha
                lda #%00011000 + ((SCREENSHRINK + 6) & 7)
                ldx #%00000111
                dey
                sta $d011
                sax $d016
                beq toprowout
-               dex
                bne -
                nop
.if NTSC
                nop
.endif
                beq withtoprow; jmp

.if SCREENSHRINK >= 50
linecrunchentry ora #%01011000
                sta $d016
                lda #$82
                sta $dd0d
                lda #BG
                sta $d021
                sta $d022
                tsx
                stx STACKPOINTER
                ldx #$3e
                txs
                lda #%00010000
                sta $d018
.else
linecrunch      jsr delay
                nop
                tsx
                stx STACKPOINTER
                lda #%00010000
                sta $d018
                lda #$82
                sta $dd0d
                lda $d011
                adc #1
linecrunchentry clc
                and #7
                ora #%01011000
                sta $d011
                sta $d016
                lda #BG
                dey
                bne linecrunch
                sty $d015
                sta $d021
                sta $d022
                nop
                ldx #$3e
                txs
.endif
                ldy SCROLLY
                bne withtoprow
                .cerror (>*) != (>withtoprow), "irqhandler crosses page boundary"

                nop
                nop
toprowout       ldx #%01010001
                lda YSCROLLTOP
                bmi torowheight8

                stx $dd0f
                sta $d016
                sta $d011
                sty $d018
                ldy #(>COLUMNSELECT)
                jmp COLUMNSELECT

torowheight8    stx $dd0f
                sta $d016
                sta $d011
                sty $d018
                ldx XSCROLL
                bpl torowheight9pls
                lda #%01010111
                sax $d016
                sta $dd0f
                lda YSCROLLBOTTOM
                sta $d011
                ldy #(>COLUMNSELECT)
                jmp COLUMNSELECT

torowheight9pls ldy #(>COLUMNSELECT)
                jmp rowheight9plus

irqout          asl $dd0f

                ldx FILTERSWEEP
                lda FILTERVALLO
                asl FILTERDIR
                bcs +

                adc filtspeedlo,x
                tay
                lda FILTERVALHI
                adc filtspeedhi,x
                cmp filttop,x
                bvc ++; jmp

+               sbc filtspeedlo,x
                tay
                lda FILTERVALHI
                sbc filtspeedhi,x
                cmp #2

+               ror FILTERDIR
                sty FILTERVALLO
                sty SID_FILTER_CUTOFF_LO
                sta FILTERVALHI
                sta SID_FILTER_CUTOFF_HI

                sec
                ldx #14

voiceloop       dec DURATION,x
                bpl soundframe

                lda #FRAMESPERTICK - 1
                sta DURATION,x
                ldy SEQPOS,x
                bne getnote

nextseq        ;sec
                lda TRACKPOS,x
                adc #2 - 1
                and #%00001111
                sta TRACKPOS,x

                ldy trackseq,x

getnote         lda seqdata,y
                beq nextseq

                bpl instrumnornote

                sta FILTERDIR
                alr #%01111111
                tax
                stx FILTERSWEEP
                lda filttop,x
                sta FILTERVALHI
                bcs +
                lda #(2 * FRAMESPERTICK) - 1
                sta DURATION + 0
+               ldx #0
                iny
                lda seqdata,y

instrumnornote  lsr
                bcc +
                sta INSTR,x
                iny
                lda seqdata,y
                lsr

+              ;clc
                iny
                sty SEQPOS,x

                ldy TRACKPOS,x
                adc tracks,y
                sta NOTE,x

                ldy INSTR,x
                lda adtable,y
                sta SID_V1_AD,x
                lda srtable,y
                sta SID_V1_SR,x
                lda pulsehilim,y
                sta PULSEHICOMP,x
                sta PULSEHI,x
                lda pulseadd,y
                sta PULSEADDLO,x
                lda instruments,y
                sta INSTRBASE,x
                lda #9
                sta SID_V1_CONTROL,x
                lda #0
                beq drumfreq; jmp

wavetablejump   lda wavetable + 1,y
                sta INSTRBASE,x

soundframe     ;sec
                bne modulate

                anc #0
                sta PULSEADDHI,x
                sta PULSELO,x
                sta SID_V1_SR,x

modulate        ldy INSTRBASE,x
                lda wavetable,y
                beq wavetablejump

                inc INSTRBASE,x
                bcs setwave
                and #$fe
setwave         sta SID_V1_CONTROL,x

                lda #$80
                cpy #wave3 - wavetable
                bcc +
                clc
                lda wavefreq - (wave3 - wavetable),y
                bpl drumfreq
               ;clc
+               adc NOTE,x
                ldy #$01
                sty FREQHI
               ;clc
                sbc #24 - 1
                bpl +
-               iny
               ;sec
                sbc #12
                bmi -
+               sty SCRATCH
                tay
                lda notefreqs - $80 + 12,y
                ldy SCRATCH
                dey
                bmi +
-               asl
                rol FREQHI
                dey
                bpl -
+               adc DETUNE,x
                sta SID_V1_FREQ_LO,x
                lda FREQHI
                adc #0
drumfreq        sta SID_V1_FREQ_HI,x

               ;clc
                lda PULSEADDLO,x
                adc PULSELO,x
                sta PULSELO,x
                sta SID_V1_PULSE_LO,x
                lda PULSEADDHI,x
                adc PULSEHI,x
                sta PULSEHI,x
                sta SID_V1_PULSE_HI,x

                cmp PULSEHICOMP,x
                ldy #0
                bcc +
                dey
                eor #%00001111
                cmp PULSEHICOMP,x
                bne ++
+               lda #$ff
                eor PULSEADDLO,x
                sta PULSEADDLO,x
                sty PULSEADDHI,x

+               txa
                sbx #7
                bcc update
                jmp voiceloop

update          asl REMAPSWITCH
                beq ++
                REMAPOFFS = 11
                ldy rowmapping + REMAPOFFS
                lda ZOOMROWMAPPING
                sty ZOOMROWMAPPING
                sta rowmapping + REMAPOFFS
                ldy rowmapping + REMAPOFFS + 1
                lda ZOOMROWMAPPING + 1
                sty ZOOMROWMAPPING + 1
                sta rowmapping + REMAPOFFS + 1
                ldy rowmapping + REMAPOFFS + 2
                lda ZOOMROWMAPPING + 2
                sty ZOOMROWMAPPING + 2
                sta rowmapping + REMAPOFFS + 2
                ldx #0
                stx REMAPSWITCH
                bcs +
                stx ZOOMSWITCH
                stx SCALE
                stx SCALE + 1
                inx
+               stx YSCALESWITCH
+
                lda #>(renderchunks.chunk[COLUMNS - 1])
                jsr setdesthibyte

                ldy SCALE
                lax SCALE + 1
                sty xscalelo
                sta xscalehi
                sbx #-(3 * 8)
                sty centerscalelo
                stx centerscalehi
                lda SCROLL
                jsr mulu_16_8_16

                sta SOFTSCROLL

                CENTERCHAR = 26
                ldx #0
                stx CENTERHI
                ldy #$0100 - (CENTERCHAR * 8)
-               txa
                sbx #-(2 * CHARBUFNUMROWS)
                bcc +
                inc CENTERHI
                clc
+               lda SCRATCH
centerscalelo = * + 1
                adc #0
                sta SCRATCH
                tya
centerscalehi = * + 1
                adc #0
                tay
                bcc -
+
               ;sec
                lda SCALE
                sbc SCRATCH
                sta COUNTER
                stx CENTER
                bcs +
                iny
+               tya
                clc
                adc SOFTSCROLL
                bcs +

                cmp centerscalehi
                bcc softscrlinbound
+               sbc centerscalehi
                tay
                lax CENTER
                sbx #2 * CHARBUFNUMROWS
                stx CENTER
                bcs +
                dec CENTERHI
+               tya

softscrlinbound sta SOFTSCROLL
                sec
                lda HARDSCROLLLO
                sbc CENTER
                sta SOURCE
                lda HARDSCROLLHI
                sbc CENTERHI
                cmp #>CHARBUFFER
                bcs +
                adc #CHARBUFNUMPAGES
+               jsr setsourcehi
                lax centerscalehi
                sbx #-$f8
                ldx #<(renderchunks.chunk[COLUMNS - 3])
                bcc entrynostretch

                sbc SOFTSCROLL
                bcc +
                tay
                bcs stretchloop; jmp

entrynostretch  sec
                sbc SOFTSCROLL
+               tay
                ora #$f8
                sta COUNTER + 1
                cpy #$e8
                bcs countergeqe8; nostretch

               ;clc
                lda #(2 * CHARBUFNUMROWS) - 1
                adc SOURCE
                sta SOURCE
                lda #0
                jsr doincsourcehi
                ldy COUNTER + 1
                jmp aftdecdestloop - 1; jmp

countergeqe8    cpy #$f8
                bcs nostretchloop
                ldx #<(renderchunks.chunk[COLUMNS - 2])
                cpy #$f0
                bcs +
                ldx #<(renderchunks.chunk[COLUMNS - 1])
+               lda COUNTER + 1
                tay
                anc #%00000111
                bcc nostretchpart; jmp

-               jsr incdesthibyte
                sec
                bcs +; jmp

stretchloop     lda #EMPTYCHAR
dest0           sta $ff00 + MASKOFFSET + (2 * RENDERCHUNKSIZE),x
                lda SOURCE
dest1           sta $ff00 + CHARBUFFEROFFSET + (2 * RENDERCHUNKSIZE),x
source1 = * + 1
                lda #0
dest2           sta $ff00 + CHARBUFFEROFFSET + 1 + (2 * RENDERCHUNKSIZE),x
                txa
                sbx #RENDERCHUNKSIZE
                bcc -
+               tya
               ;sec
                sbc #8
                tay
                bcs stretchloop

nostretchloop   anc #%00000111
dest3           sta $ff00 + MASKOFFSET + (2 * RENDERCHUNKSIZE),x

nostretchpart
dest4           sta $ff00 + MASKOFFSET + RENDERCHUNKSIZE,x
                ora #%00010000
dest5           sta $ff00 + MASKOFFSET,x
source2 = * + 1
                lda #0
dest6           sta $ff00 + CHARBUFFEROFFSET + 1 + (2 * RENDERCHUNKSIZE),x
dest7           sta $ff00 + CHARBUFFEROFFSET + 1 + (1 * RENDERCHUNKSIZE),x
dest8           sta $ff00 + CHARBUFFEROFFSET + 1,x
                lda SOURCE
dest9           sta $ff00 + CHARBUFFEROFFSET + (2 * RENDERCHUNKSIZE),x
               ;clc
                adc #CHARBUFNUMROWS
dest10          sta $ff00 + CHARBUFFEROFFSET + RENDERCHUNKSIZE,x
dest11          sta $ff00 + CHARBUFFEROFFSET,x
               ;clc
                adc #CHARBUFNUMROWS
                sta SOURCE
                bcs incsourcehi

makemaskloop    txa
                sbx #3 * RENDERCHUNKSIZE
                bcc decdestloop
                clc
aftdecdestloop  lda COUNTER
xscalelo = * + 1
                adc #0
                sta COUNTER
                tya
xscalehi = * + 1
                adc #0
                tay
                bcc nostretchloop
                bcs stretchloop; jmp
                .cerror (>*) != (>nostretchloop), "maskloop crosses page boundary"

incsourcehi    ;sec
               ;lda #0
                adc source1
                jsr incwrapcharbufh
                sta source1
                sta source2
                jmp makemaskloop

decdestloop     jsr incdesthibyte
                bcc aftdecdestloop; jmp

incdesthibyte   lda #>(renderchunks.chunk[0])
                cmp dest0 + 2
                beq generatey
setdesthibyte   sta dest0 + 2
                sta dest1 + 2
                sta dest2 + 2
                sta dest3 + 2
                sta dest4 + 2
                sta dest5 + 2
                sta dest6 + 2
                sta dest7 + 2
                sta dest8 + 2
                sta dest9 + 2
                sta dest10 + 2
                sta dest11 + 2
delay           rts

generatey       jsr scaley
                sta yscalelimithi
                sta YSCALEHI
                lda SCRATCH
                sta yscalelimitlo
                sta YSCALELO

                ldx #>COLUMNSELECT
                stx columnselstore + 2
                dex
                stx rowselectstore + 2
                lda #<(rowmapping + 9 + 3)
                sta ROWCOUNT
                sec
                lda #$80
                sta $d021
                sta $d022
                sbc YSCALELO
                tay
                lda #192 - SCREENSHRINK - 1
                sbc YSCALEHI
                lsr
                tax
                tya
                ror
                tay
                sec
                txa
-               tax
                tya
yscalelimitlo = * + 1
                sbc #0
                tay
                txa
yscalelimithi = * + 1
                sbc #0
                dec ROWCOUNT
                bcs -
                tya
                adc yscalelimitlo
                sta SCRATCHHI
                lda ROWCOUNT
                cpx #7
                ldy #<renderchunks.chunk[1]
                bcs notopout

                sta secondtoprowcnt + 1
                adc #1
                sta rowselect
                sta toprowcount + 1
                stx SCROLLY
                cpx #SCREENSHRINKMOVE
                bcc +
                ldy #<renderchunks.chunk[2]
+               sty COLUMNCOUNT
                sty toprowentry + 1
                txa
                axs #-($37 + 7 + SCREENSHRINK)
                ldy #0
               ;clc
                bcc generateyentry; jmp

notopout        sta rowselect
                sta toprowcount + 1
               ;ldy #<renderchunks.chunk[1]
                sty COLUMNCOUNT
                sty toprowentry + 1
                txa
                ldx #$37 + 7 + SCREENSHRINK
                ldy #0
                sty SCROLLY
                clc
                bcc generateyrow; jmp

generateystrlim cmp #7 + (7 + 6)
                bcs +
                lda #7
                bcc generateystlmc; jmp
+               lda #7 + 6
                bne generateystlmc; jmp

genheight7noclm ora #%01000000
                sta YSCROLLTOP,y
                lda COLUMNCOUNT
                bcc columnselstore; jmp

generateynext   iny
                tax

                inc columnselstore + 2
                inc rowselectstore + 2

                sec
                lda SCRATCH
                sbc ROWHEIGHT
                bne generateyrow
                inc rowselect

generateyentry ;clc
                lda YSCALELO
                adc SCRATCHHI
                sta SCRATCHHI
                lda YSCALEHI
                adc #0

generateyrow    sta SCRATCH
                cmp #7 + 7
                bcs generateystrlim
generateystlmc  sta ROWHEIGHT
                cmp #8
                beq genheight8

                txa
                and #%00000111
                bcs genheight9plus
                beq genheight7noclm

                ora #%01000000
                sta YSCROLLTOP,y
                bcc nextrow; jmp

genheight9plus  sta XSCROLL,y
                ora #%11000000
                sta YSCROLLTOP,y
               ;sec
                adc ROWHEIGHT
                adc #%00110000
                anc #%01000111
                sta YSCROLLBOTTOM,y
               ;clc
                sbc #0
                cmp #$40
                lda #2 * RENDERCHUNKSIZE
                bcc +
                lsr
                bcc +; jmp

genheight8      txa
                and #%00000111
                ora #%11000000
                sta YSCROLLTOP,y
               ;sec
                adc #0
                sta XSCROLL,y
                adc #%00111000
                anc #%01000111
                sta YSCROLLBOTTOM,y
nextrow         lda #RENDERCHUNKSIZE
+               adc COLUMNCOUNT
                sta COLUMNCOUNT
columnselstore  sta $ff03; jmp lobyte
rowselect = * + 1
                lda rowmapping
rowselectstore  sta $ff01; ldx #imm argument

               ;clc
                txa
                adc ROWHEIGHT
                bcc generateynext
                .cerror (>(*)) != (>generateynext), "generatey crosses page boundary"

                lda #%00000100
                sta $d015

                ; limit bottom row to 7 pixels if it starts on rasterline >= $f8
                cpx #$f8
                bcc +
                clc
                txa
                eor #%10111000
                sta YSCROLLTOP,y
+
                TOPROWCHUNK = 41
                lda renderchunks.chunk[TOPROWCHUNK] + 4
                sta SCRATCH
                lda #$4c; jmp
                sta renderchunks.chunk[TOPROWCHUNK] + 3
                lda #>secondtoprowout
                sta renderchunks.chunk[TOPROWCHUNK] + 5

                lda SCROLLY
                beq generatetoprow

                ldx #$3e
                txs
                lda #<secondtoprowout
                sta renderchunks.chunk[TOPROWCHUNK] + 4
secondtoprowcnt ldx rowmapping
                jmp renderchunks.chunk[0]

secondtoprowout ldx #9
-               lda $0118,x
                sta $0518,x
                lda $0118 + 10,x
                sta $0518 + 10,x
                lda $0118 + 20,x
                sta $0518 + 20,x
                lda $0118 + 30,x
                sta $0518 + 30,x
                dex
                bpl -

generatetoprow  ldx #$3e
                txs
                lda #<toprowout3
                sta renderchunks.chunk[TOPROWCHUNK] + 4
toprowcount     ldx rowmapping
toprowentry     jmp renderchunks.chunk[1]

toprowout3      lda #$09; ora #imm
                sta renderchunks.chunk[TOPROWCHUNK] + 3
                lda SCRATCH
                sta renderchunks.chunk[TOPROWCHUNK] + 4
                lda #$48; pha
                sta renderchunks.chunk[TOPROWCHUNK] + 5

                .cerror (>secondtoprowout) != (>toprowout3), "toprowout page mismatch"

               ;clc
                lda #SOFTSCROLLSPEED
                adc SCROLL
                sta SCROLL
                bcc +
                lda #(2 * CHARBUFNUMROWS) - 1
                adc HARDSCROLLLO
                sta HARDSCROLLLO
                lda HARDSCROLLHI
                jsr incwrapcharbufh
                sta HARDSCROLLHI
                inc SCROLLSEMAPHORE
+
                lda #<ZOOMSPEED
                ldy #>ZOOMSPEED
                ldx ZOOMSWITCH
                bne +
                ldy #0
                lda SLOWZOOM
                bpl +
                dey
+               clc
                adc SCALE
                tax
                tya
                adc SCALE + 1
                tay
                iny
                beq +
                cmp #MAXZOOM
                bcc ++
                lda #MAXZOOM
                .byte $a0; ldy #imm
+               tya
                ldx #0
+               stx SCALE
                sta SCALE + 1

                lda columnselstore + 2
                sta * + 7
                ldx #0
                stx COLUMNSELECT + 3; ldx #imm argument

                inc $dc01
                bne exit

                ldx STACKPOINTER
                txs
                lda #$31
                sta $d019
updatedone      rts
                .byte $01
                pla
                tay
                pla
                tax
                pla
                rti

exit            stx $d011

-               lda LOWMEM0BUFFER,x
                sta LOWMEM0,x
                lda logosprite,x
                sta LOWMEM1,x
                inx
                bne -

                ldx #13
-               ldy #$0100 - RENDERJUMPSIZE
restorepayload  lda CHARBUFFER,x
                sta COLUMNSELECT - $0200 + RENDERJUMPSIZE,y
                inx
                bne +
                inc restorepayload + 2
+               txa
                bit CONST0F
                bne +
                sbx #-13
+               iny
                bne restorepayload
                inc restorepayload + 5
                lda restorepayload + 5
                cmp #(>COLUMNSELECT) + MAXROWS - 1
                bne -; jmp

                inc $01
                jmp ($fffc)

incwrapcharbufh adc #0
                cmp #(>CHARBUFFER) + CHARBUFNUMPAGES
                bcc +
                sbc #CHARBUFNUMPAGES
+               rts

doincsourcehi   adc source1
                jsr incwrapcharbufh
setsourcehi     sta source1
                sta source2
                rts

run             sty $d803 + LOGOROWOFFSET
                stx $d023
                stx $d024

                ldy #$1f + SCREENSHRINK
                sty $d005
                ldy #$4c
                sty $dd0c
-
.for charbufferpage := CHARBUFNUMPAGES - 1, charbufferpage < CHARBUFNUMPAGES, charbufferpage += 1
                sta CHARBUFFER + ($0100 * charbufferpage),x
.next
                inx
                bne -

               ;lda #$18
                sta $d004

                lda #$0540 / $40
                sta $07fa

                ldx #13
-               ldy #$0100 - RENDERJUMPSIZE
-               lda (POINTER),y
bufferpayload   sta CHARBUFFER,x
                lda renderjump - $0100 + RENDERJUMPSIZE,y
                sta (POINTER),y
                inx
                bne +
                inc bufferpayload + 2
+               txa
                bit CONST0F
                bne +
                sbx #-13
+               iny
                bne -
                inc POINTER + 1
                lda POINTER + 1
                cmp #(>COLUMNSELECT) + MAXROWS - 1
                bne --; jmp

                inc TRACKPOS
                inc DURATION + ((OFFBEATVOICE - 1) * 7)
                lda #DETUNEADD
                sta DETUNE + ((DETUNEDVOICE - 1) * 7)
                lda #$f0 + FILTERVOICE
                sta SID_FILTER_CONTROL_1
                lda #FILTERTYPE + $0f
                sta SID_FILTER_CTRL2_VOL

                ldx #LOOPTOS
                txs
                jsr update
                ldx #$85; sta zp
                stx updatedone
                sta $d01a

mainloop        lda SCROLLSEMAPHORE
                beq mainloop

                lda #EMPTYCHAR
                ldx #6
-               ldy topbottomclear - 1,x
                sta (CHARBUFFERPTR),y
                dex
                bne -

-               stx CHARPTR
                asl CHARSCRATCH + 8,x
                lda CHARSCRATCH,x
                rol
                and #%00000011
                sta CHARSCRATCH,x
                tax
                lda LEFTTABLE,x
                ldy #0 + 2
                sta (CHARBUFFERPTR),y
                ldy #CHARBUFNUMROWS + 2
                lda RIGHTTABLE,x
                sta (CHARBUFFERPTR),y
                inc CHARBUFFERPTR
                ldx CHARPTR
                inx
                cpx #8
                bne -

               ;sec
                lda #(2 * CHARBUFNUMROWS) - 8 - 1
                adc CHARBUFFERPTR
                sta CHARBUFFERPTR
                lda CHARBUFFERPTR + 1
                jsr incwrapcharbufh
                sta CHARBUFFERPTR + 1

                dec SCROLLSEMAPHORE

                dec CHARCOLUMN
                bpl mainloop

                lda ZOOMSWITCH
                bne +
                inc SCROLLPOS
+               jsr makecharbuffer
                bvc mainloop; jmp

rowmapping      .byte 11, 11, 11, 11, 11, 11, 11, 11
                .byte 1
                .byte 2, 3, 4, 5, 6, 7, 8, 9
                .byte 0, 0, 0, 0, 0, 0, 0, 0, 0
                .cerror (>*) != (>rowmapping), "rowmapping crosses page boundary"

topbottomclear = * - 1
                .byte 1, 11, 16, 1 + 16, 11 + 16

-               sta SCROLLPOS
makecharbuffer  ldx SCROLLPOS
                lda text,x
                beq -
                cmp #' '
                beq zoom
                pha
                asl
                rol
                rol
                ldx #%11111000
                sax CHARCOLUMN
                rol
                and #%00000111
                ora #>$d800
                sta CHARCOLUMN + 1
                ldy #7
getchar         lda (CHARCOLUMN),y
                sta CHARSCRATCH + 8,y
                dey
                bpl getchar
                ldx #3
                pla
                cmp #'"'
                beq +

                ldx #6
                lda CHARSCRATCH + 8 + 7
-               ora CHARSCRATCH + 8,x
                dex
                bpl -
-               ldx #8
-               asl CHARSCRATCH + 8 - 1,x
                dex
                bne -
                asl
                bpl --
-               inx
                asl
                bne -
+               stx CHARCOLUMN
-               rts

zoom           ;sec
                ror ZOOMSWITCH

                lda SCALE + 1
                eor #MAXZOOM
                bne -

                ldx HARDSCROLLLO
                stx ZOOMSRCLO
                ldx HARDSCROLLHI
                stx ZOOMSRCHI

               ;lda #0
                ldx #6
                jsr genzoom
                lda #11
                ldx #4
                jsr genzoom
                lda #12
                ldx #5
                jsr genzoom

                inc REMAPSWITCH

-               ldx REMAPSWITCH
                bne -

                ldy #>CHARBUFFER
-               sty copyzoom + 2
                sty copyzoom + 7
               ;ldx #0
copyzoom        lda CHARBUFFER + 12,x
                ldy #9
-               sta CHARBUFFER + 1,x
                inx
                dey
                bne -
                txa
                axs #-(CHARBUFNUMROWS - 9)
                bcc copyzoom
                ldy copyzoom + 2
                iny
                cpy #>(CHARBUFFER + ($0100 * CHARBUFNUMPAGES))
                bcc --

                dec REMAPSWITCH

                lda #SLOWZOOMSPEED
                sta SLOWZOOM
                rts

genzoom         sta putzoom + 1
                txa
                clc
                ora #<(($0100 * CHARBUFNUMPAGES) - (2 * CHARBUFNUMROWS))
                adc ZOOMSRCLO
                sta getzoom + 1
                and #%11110000
                tax
                lda #>(($0100 * CHARBUFNUMPAGES) - (2 * CHARBUFNUMROWS))
                adc ZOOMSRCHI
                jsr incwrapcharbufh
                sta getzoom + 2
                clc
                adc #>(($0100 * CHARBUFNUMPAGES) - (2 * CHARBUFNUMROWS))
                jsr incwrapcharbufh
                sta putzoom + 2

                lda #CHARBUFNUMPAGES * CHARBUFNUMROWS
                sta ZOOMCOUNT
getzoom         lda CHARBUFFER
                ldy #CHARBUFNUMROWS + 2
                cpy ZOOMCOUNT
                bcc putzoom
                ldy ZOOMCOUNT
                bcs putzoom; jmp

-               cmp #%01001000
                bcc putzoom

               ;cmp #FULLTOEMPTYRIGHTCHAR
                bne +
                lda #EMPTYCHAR
                bne putzoom; jmp

+               cmp #EMPTYTOFULLLEFTCHAR
                bne +
                lda #EMPTYTOFULLLRIGHTCHAR
                bne putzoom; jmp

+               cmp #EMPTYTOFULLLRIGHTCHAR
                bne +
                lda #FULLCHAR
                bne putzoom; jmp

+              ;cmp #FULLTOEMPTYLEFTCHAR
               ;bne +
                lda #FULLTOEMPTYRIGHTCHAR

putzoom         sta CHARBUFFER,x
                pha
                txa
                sbx #-CHARBUFNUMROWS
                lda putzoom + 2
                jsr incwrapcharbufh
                sta putzoom + 2
                pla
                dey
                bne -

                sec
                lda ZOOMCOUNT
                sbc #CHARBUFNUMROWS + 2
                bcc zoomout
                sta ZOOMCOUNT

                lda #(2 * CHARBUFNUMROWS) - 1
                adc getzoom + 1
                sta getzoom + 1
                lda getzoom + 2
                jsr incwrapcharbufh
                sta getzoom + 2
                bne getzoom; jmp

scaley
.if SCREENSHRINK >= 50
                lda #%01011000 + ((SCREENSHRINK + 6) & 7)
.else
                lda #%01011000 + ((SCREENSHRINK + ((SCREENSHRINK + 3) / 8) - 1) & 7)
.endif
                sta $d011
                lda #%00001000 + 4
                sta $d016
                lda #%00010110
                sta $d018

                ldx centerscalehi
                ldy YSCALESWITCH
                beq +
                ldx #MAXZOOM + (3 * 8)
+               ldy SCALE
                lda #TEXELASPECTRATIO

mulu_16_8_16    sty factorlo
                stx factorhi
                sec
                ror
                sta SCRATCHHI
                lax #0
                sta SCRATCH
-               bcc +
                clc
                tax
                lda SCRATCH
factorlo = * + 1
                adc #0
                sta SCRATCH
                txa
factorhi = * + 1
                adc #0
+               ror
                ror SCRATCH
                lsr SCRATCHHI
                bne -
zoomout         rts

SID_V1_FREQ_LO       = $d400
SID_V1_FREQ_HI       = $d401
SID_V1_PULSE_LO      = $d402
SID_V1_PULSE_HI      = $d403
SID_V1_CONTROL       = $d404
SID_V1_AD            = $d405
SID_V1_SR            = $d406
SID_FILTER_CUTOFF_LO = $d415
SID_FILTER_CUTOFF_HI = $d416
SID_FILTER_CONTROL_1 = $d417
SID_FILTER_CTRL2_VOL = $d418

.if NTSC
FRAMESPERTICK   = 8
.else
FRAMESPERTICK   = 7
.endif
OFFBEATVOICE    = 2

DETUNEDVOICE    = 3
DETUNEADD       = $1e

FILTERVOICE     = 1
FILTERTYPE      = $30

DURATION        = MUSICZP + 0
TRACKPOS        = MUSICZP + 1
SEQPOS          = MUSICZP + 2
DETUNE          = MUSICZP + 3
PULSELO         = MUSICZP + 4
PULSEHI         = MUSICZP + 5
PULSEHICOMP     = MUSICZP + 6

MUSICZP1        = MUSICZP + 21
INSTRBASE       = MUSICZP1 + 2
INSTR           = MUSICZP1 + 3
NOTE            = MUSICZP1 + 4
PULSEADDLO      = MUSICZP1 + 5
PULSEADDHI      = MUSICZP1 + 6

TUNEZP          = $40
FILTERDIR       = TUNEZP + $07
FREQHI          = TUNEZP + $0f
FILTERSWEEP     = TUNEZP + $17
FILTERVALLO     = TUNEZP + $1f
FILTERVALHI     = TUNEZP + $27

tunedata

tracks          .byte 15
                .byte 2

                .byte 0
                .byte 6
                .byte 0
                .byte 6
                .byte 10
                .byte 0
                .byte 12
                .byte 0
                .byte 0
                .byte 6
                .byte 0
                .byte 6
                .byte 12
                .byte 2

seqdata
                in0 = 1
                in1 = 3
                in2 = 5
                in3 = 7
                in4 = 9
                in5 = 11

                fs0dur = $80
                fs0    = $81
                fs1    = $83
                fs2    = $85

                endmark = $00

seq1            .byte fs0dur
                .byte e + o1
                .byte fs1
                .byte e + o2
                .byte fs2, in5
                .byte cx + o3
                .byte fs0dur, in0
                .byte e + o1
                .byte fs2, in5
                .byte cx + o3
                .byte fs0, in0
                .byte e + o1
                .byte endmark

seq2            .byte in1
                .byte ax + o3
                .byte in2
                .byte cx + o4
                .byte in1
                .byte fx + o4
                .byte cx + o4
                .byte in2
                .byte cx + o3
                .byte in1
                .byte d + o4
                .byte in2
                .byte d + o3
                .byte in1
                .byte cx + o4
                .byte endmark

instruments = * - 1
               ;.byte wave0 - wavetable
                .byte wave3 - wavetable
                .byte wave1 - wavetable
                .byte wave2 - wavetable
                .byte wave4 - wavetable
                .byte wave5 - wavetable

seq3            .byte in4
                .byte dx + o2
                .byte in2
                .byte cx + o4
                .byte in3
                .byte ax + o2
                .byte in2
                .byte ax + o3
                .byte in3
                .byte dx + o2
                .byte in2
                .byte fx + o4
                .byte in3
                .byte cx + o3
                .byte in2
                .byte d + o4
                .byte endmark

                track1seq = seq1 - seqdata
                track2seq = seq2 - seqdata
                track3seq = seq3 - seqdata

trackseq = * - 1
               ;.byte track1seq

adtable         .byte $00, $04, $00, $69, $0f, $0f

                .byte track2seq

srtable         .byte $ff, $49, $89, $7d, $f9, $f5

                .byte track3seq

pulsehilim      .byte $01, $07, $08, $01, $08, $08
pulseadd        .byte $7e, $44, $00, $f0, $08, $08

                jump = $00
wavetable
wave0           .byte $41
                .byte $41
                .byte $41
                .byte $41
                .byte $41
                .byte $41
                .byte $41
-               .byte $55
                .byte jump
                .byte (-) -wavetable

wave1
-               .byte $11
                .byte jump
                .byte (-) -wavetable

wave2
-               .byte $45
                .byte jump
                .byte (-) -wavetable

wave3           .byte $41
-               .byte $40
                .byte jump
                .byte (-) -wavetable

wave4           .byte $81
                .byte $41
                .byte $41
                .byte $41
                .byte $11
-               .byte $10
                .byte jump
                .byte (-) -wavetable

wave5           .byte $81
                .byte $41
                .byte $41
                .byte $54
-               .byte $54
                .byte $80
                .byte jump
                .byte (-) -wavetable

                F = $80
wavefreq        .byte $0c ^ F
                .byte $00 ^ F
                .byte   0
                .byte   0

                .byte $d8 ^ F
                .byte $8c ^ F
                .byte $88 ^ F
                .byte $86 ^ F
                .byte $84 ^ F
                .byte $82 ^ F
                .byte   0
                .byte   0

                .byte $d8 ^ F
                .byte $92 ^ F
                .byte $8e ^ F
                .byte $ba ^ F
                .byte $92 ^ F
                .byte $d8 ^ F

                M = 2

               ;c  = 1 * M
                cx = 2 * M
                d  = 3 * M
                dx = 4 * M
                e  = 5 * M
               ;f  = 6 * M
                fx = 7 * M
               ;g  = 8 * M
               ;gx = 9 * M
               ;a  = 10 * M
                ax = 11 * M
               ;b  = 12 * M

               ;o0 = 0 * 12 * M
                o1 = 1 * 12 * M
                o2 = 2 * 12 * M
                o3 = 3 * 12 * M
                o4 = 4 * 12 * M
               ;o5 = 5 * 12 * M
               ;o6 = 6 * 12 * M
               ;o7 = 7 * 12 * M

                .cerror (>*) != (>tunedata), "tunedata crosses page boundary"
logosprite
                .logical $0540

                .byte $00, $08, $00
                .byte $00, $98, $00
                .byte $01, $b8, $00
                .byte $03, $ff ^ %00000000, $80 ^ %00000000
                .byte $0f, $ff ^ %00000001, $e0 ^ %00000000
                .byte $1f, $ff ^ %00000011, $f0 ^ %10000000
                .byte $1f, $ff ^ %00000110, $f8 ^ %11000000
                .byte $3f, $ff ^ %00001100, $fc ^ %01100000
                .byte $3f, $ff ^ %00011000, $fc ^ %00110000
                .byte $7f, $ff ^ %00111111, $fe ^ %11111000
                .byte $7f, $ff, $fe
                .byte $7f, $ff, $fe
                .byte $7e, $3b, $9c
                .byte $38, $3b, $88
                .byte $10, $11, $08
                .byte $08, $00, $08
                .byte $0b, $00, $10
                .byte $05, $80, $10
                .byte $02, $78, $20
                .byte $01, $80, $c0
                .byte $00, $7f, $00

filttop         .byte $40, $4a, $70

                .byte 0

rowirqhandler   ldx #$3e
                txs
                iny
                lda YSCROLLTOP - >COLUMNSELECT,y
.if NTSC
                nop
                nop
.endif
                sta $d016
                sta $d011
                bpl rowheight7

                ldx XSCROLL - >COLUMNSELECT,y
                bpl rowheight9plus
                lda #%01010111
                sax $d016
                sta $dd0f
                lda YSCROLLBOTTOM - >COLUMNSELECT,y
                sta $d011
rowheight7      sty renderjmp8 + 2
renderjmp8      jmp COLUMNSELECT

rowheight9plus ;clc
                sbc YSCROLLBOTTOM - >COLUMNSELECT,y
               ;sec
                adc #7 - 1
                and #7
                sty renderjmp + 2
-               cmp $dd06
                bne -
                lda #%01010111
                sta $dd0f
                txa
                adc YSCROLLBOTTOM - >COLUMNSELECT,y
                sbc YSCROLLTOP - >COLUMNSELECT,y
                anc #7
                sta $d016
                lda YSCROLLBOTTOM - >COLUMNSELECT,y
                sta $d011
renderjmp       jmp COLUMNSELECT

notefreqs       .byte $06, $16, $26, $38, $4c, $5e, $74, $8a, $a2, $ba, $d4, $f0

filtspeedlo     .byte <$0128, <$0120, <$0204
filtspeedhi     .byte >$0128, >$0120, >$0204

                .here
lowmemcodeend
                .enc "screen"

text            .text ' Plush feat Triad presents "Softwired'
                .text "' with tunage by Dalezy and code by Krill now again"
                .byte 0
end
                .cerror * > PAYLOAD, "code too large"
